
/*
 * Prolific NOR flash DMA controller for MTD
 *
 * Gary Lin. July 29, 2006
 * Jun Chen, Aug. 22, 2006 (Modify for PL1061D)
 * Jun Chen, Apr. 02, 2007 (Modify for PL1029)
 */
 
#define USE_POLLING 0	// use polling instead of interrupt ?

/****************************/
/*  Choose 1 Nor Flash Chip */
/****************************/
#define USE_PPI_EN29LV320B 1	
#define USE_SPI_S25FL032A  0
#define USE_PPI_INTEL_P30  0
/****************************/
#define NORMTD_VERSION "1.0"
 
#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/module.h>
#include <linux/slab.h>

#include <asm/io.h>
#include <asm/uaccess.h>
//#include <asm/pl1061/dma.h>
#include <asm/dma.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/setup.h>
#include <asm/semaphore.h>

#include <linux/mtd/mtd.h> 
#include <linux/mtd/partitions.h>

#include <linux/mtd/nor.h> 

#ifdef CONFIG_PL_NOR_PPI_AMD
#include <linux/mtd/nor_ppi_amd.h> 
#endif

#ifdef CONFIG_PL_NOR_PPI_INTEL
#include <linux/mtd/nor_ppi_intel.h> 
#endif

#ifdef CONFIG_PL_NOR_SPI
#include <linux/mtd/nor_spi.h> 
#endif



#define MAX_NOR_SECTOR 1024
#define NOR_TYPE_NUM 3
static const char *nor_type[NOR_TYPE_NUM] = { "spi", "ppi-amd", "ppi-intel" };
static int norflash_tag = 0;
static struct tag_norinfo nor_geometry;
typedef struct {
	unsigned long start_addr;
	unsigned int size;
} nor_sector_t;
static nor_sector_t nor_sector_map[MAX_NOR_SECTOR];
static int total_sectors = 0;
static int total_size = 0; /* in bytes */
static int num_erase_regions = 0;

/*
 * Actually 4MB-1, but we want easy implementation for 16-byte aligned, so...
 */
#define MAX_DMA_CNT	0x3fff00

typedef struct {
	int (*init)(void);
	int (*reset)(void);
	int (*read_id)(u8 *);
	int (*cmd_erase)(unsigned long, int);
	int (*cmd_write)(unsigned long, int, int);
	int (*cmd_read)(unsigned long, int, int);
	int (*pio_read)(unsigned long, char *, int);
	int (*pio_write)(unsigned long, char *, int);
} pl_nor_ops_t;
static pl_nor_ops_t *nor_ops = NULL;


static u8 intr_status;
#if 0
static char *dma_buf = NULL;
static int dma_buf_size = 0;
#endif

#define MDE_NXT_SHIFT           21
#define MDE_VALID_SHIFT         30
#define MDE_EMDE_SHIFT          31

typedef struct {
	unsigned long saddr;
	unsigned long mde;
} dma_mde_t;

#ifndef MAX_SECTORS
#define MAX_SECTORS 128
#endif
#define MDE_MAX_ENTRY   MAX_SECTORS
static dma_mde_t mdt[MDE_MAX_ENTRY] __attribute__ ((aligned (8)));
static struct iovec iovec_tbl[MDE_MAX_ENTRY+1];

/*
 * taginfo passed by Proboot, it's the nor flash mapping table 
 * We rely totally on this info. Without this, we won't initialize.
 */
#define MAX_SECTOR_SET	16
struct tag_norinfo {
        u32	type;
        struct tag_sectinfo {
                u16	size;
                u16	quantum;
        } sectinfo[MAX_SECTOR_SET];
};
#define ATAG_NORINFO 0x10290101
/*
static int __init pl1029_parse_tag_norinfo(const struct tag *tag)
{
        memcpy(&nor_geometry, &(tag->u), sizeof(nor_geometry));
        norflash_tag = 1;
        
        return 0;
}
__tagtable(ATAG_NORINFO, pl1029_parse_tag_norinfo);
*/

/* Prolific DMA controller related functions */

static int pl_setup_dma(int mode, struct iovec *vecs, int count)
{
        int i, total_len;

        set_dma_mode(DMA_PL_NOR, mode);
        set_dma_mde_idx(DMA_PL_NOR, 0);
        
        /* setup MDE table */
        total_len = 0;
        for ( i=0; i<count; i++ ) {
                mdt[i].saddr = virt_to_bus(vecs[i].iov_base);
                mdt[i].mde = vecs[i].iov_len |
                             ((i + 1) << MDE_NXT_SHIFT) |
                             (1 << MDE_VALID_SHIFT);
                total_len += vecs[i].iov_len;
                pci_map_single(NULL, vecs[i].iov_base, vecs[i].iov_len,
                               mode == DMA_MODE_READ ? PCI_DMA_FROMDEVICE : 
                                                       PCI_DMA_TODEVICE);
        }
        mdt[count-1].mde |= (1 << MDE_EMDE_SHIFT);

        pci_map_single(NULL, mdt, sizeof(dma_mde_t)*(MDE_MAX_ENTRY+1), 
                       PCI_DMA_TODEVICE);

        return total_len;
}

static void pl_unmap_dma(struct iovec *vecs, int count, int mode)
{
        int i;
        
        for ( i=0; i<count; i++ ) {
                pci_unmap_single(NULL, virt_to_bus(vecs[i].iov_base),
                                 vecs[i].iov_len, 
                                 mode == DMA_MODE_READ ? PCI_DMA_FROMDEVICE : 
                                                         PCI_DMA_TODEVICE);
        }
        pci_unmap_single(NULL, virt_to_bus(mdt), 
                         sizeof(dma_mde_t)*(MDE_MAX_ENTRY+1), PCI_DMA_TODEVICE);
}

/* nor flash controller semaphore */
static DECLARE_MUTEX(ctl_sem);
static DECLARE_WAIT_QUEUE_HEAD(wq_intr);


#ifdef CONFIG_PL_NOR_SPI
pl_nor_ops_t spi = {
	.init =		spi_init,
	.reset = 	spi_reset,
	.read_id = 	spi_read_id,
	.cmd_erase = 	spi_cmd_erase,
	.cmd_write =	spi_cmd_write,
	.cmd_read =	spi_cmd_read,
	.pio_read = 	spi_pio_read,
	.pio_write =	spi_pio_write,
};
#endif

#ifdef CONFIG_PL_NOR_PPI_AMD      
pl_nor_ops_t ppi_amd = {
	.init =		ppi_amd_init,
	.reset = 	ppi_amd_reset,
	.read_id = 	ppi_amd_read_id,
	.cmd_erase = 	ppi_amd_cmd_erase,
	.cmd_write =	ppi_amd_cmd_write,
	.cmd_read =	ppi_amd_cmd_read,
	.pio_read = 	ppi_amd_pio_read,
	.pio_write =	ppi_amd_pio_write,
};
#endif

#ifdef CONFIG_PL_NOR_PPI_INTEL
pl_nor_ops_t ppi_intel = {
	.init =		ppi_intel_init,
	.reset = 	ppi_intel_reset,
	.read_id = 	ppi_intel_read_id,
	.cmd_erase = 	ppi_intel_cmd_erase,
	.cmd_write =	ppi_intel_cmd_write,
	.cmd_read =	ppi_intel_cmd_read,
	.pio_read = 	ppi_intel_pio_read,
	.pio_write =	ppi_intel_pio_write,
};
#endif

/* NOR flash controller related functions */

/* for polling the controller interrupt status registers */
int pl_nor_polling_intr(int intr_bit, int expected)
{
	unsigned long start_j = jiffies;
	int rc;
	
	rc = 0;	
	while ( (start_j + HZ) >= jiffies ) { /* max. polling time 1 sec. */
		intr_status = readb(CTRL_ST_REG);
		if ( (intr_status & 0x08) == 0 ){ /* controller time-out */
			printk("pl_nor.c: controller time-out\n");
			break;
		}
		if ( (intr_status & intr_bit) == expected ) {
			rc = 1;
			break;
		}
	}
	
	return rc;
}

static void pl_nor_do_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
        intr_status = readb(INTR_S_REG);
#if 0
        printk("NOR INTR: intr_status = 0x%02x\n", intr_status);
#endif        
        wake_up(&wq_intr);

        return;
}

static void __init pl_nor_mtd_banner(void)
{
        printk("\nPL-1029 NOR flash driver for MTD, version %s\n", NORMTD_VERSION);
}

static int __init pl_nor_init_controller(void)
{
/* 
 * no need to initialize the controller for now. 
 * Proboot has already done it. So, we only show some regs.
 */
#if 0
        printk("config = 0x%08x\n", readb(CONFIG_REG));
#endif        
        return 0;
}

static int __init pl_nor_generate_map(void)
{
/*********************************************/
/*          Fix Flash Type for Test          */
/*********************************************/
#if USE_PPI_EN29LV320B
	nor_geometry.type=1;
	nor_geometry.sectinfo[0].size=8;
	nor_geometry.sectinfo[0].quantum=8;
	nor_geometry.sectinfo[1].size=64;
	nor_geometry.sectinfo[1].quantum=63;
#elif USE_SPI_S25FL032A
	nor_geometry.type=0;
	nor_geometry.sectinfo[0].size=64;
	nor_geometry.sectinfo[0].quantum=64;
#elif USE_PPI_INTEL_P30
        nor_geometry.type=2;
        nor_geometry.sectinfo[0].size=32;
        nor_geometry.sectinfo[0].quantum=4;
        nor_geometry.sectinfo[1].size=128;
        nor_geometry.sectinfo[1].quantum=255;
        nor_geometry.sectinfo[2].size=128;
        nor_geometry.sectinfo[2].quantum=255;
        nor_geometry.sectinfo[3].size=32;
        nor_geometry.sectinfo[3].quantum=4;
#else

	#error "Choose 1 Nor Flash Chip"
#endif
/*********************************************/
        int i, j, size, sec_idx;
        u8 id;
        unsigned long addr;

        i = nor_geometry.type;        
        printk("NOR flash type: %s ", i<NOR_TYPE_NUM?nor_type[i]:"unknown");
        sec_idx = 0;
        addr = 0;
        for ( i=0; i<MAX_SECTOR_SET; i++ ) {
                size = nor_geometry.sectinfo[i].size;
                if ( size == 0 )
                        break;
                printk("%dx%d ", size, nor_geometry.sectinfo[i].quantum);
                num_erase_regions ++;
#if 0
                if ( (size<<10) > dma_buf_size )
                        dma_buf_size = (size<<10);
#endif                        
                for ( j=0; j<nor_geometry.sectinfo[i].quantum; j++ ) {
                        nor_sector_map[sec_idx].start_addr = addr;
                        nor_sector_map[sec_idx].size = (size<<10);
                        total_size += (size<<10);
                        addr += nor_sector_map[sec_idx].size;
                        sec_idx ++;
                        if ( sec_idx == MAX_NOR_SECTOR ) {
                                printk("NOR flash sector > %d, ignore the rest\n", MAX_NOR_SECTOR);
                                break;
                        }
                }                                 
        }
        total_sectors = sec_idx;
        printk("\n");
#if 0
for ( i=0; i<total_sectors; i++ ) {
        printk("sector %d: %lu, %u\n", i, nor_sector_map[i].start_addr, nor_sector_map[i].size);
}
#endif        
        switch ( nor_geometry.type ) {
                case 0: /* spi */
			#ifdef CONFIG_PL_NOR_SPI                
                        	nor_ops = &spi;
                        	break;
                        #else
	                        printk("NOR: not support spi flash type\n");
	                        return -1;                   
                        #endif
                case 1: /* ppi amd */
			#ifdef CONFIG_PL_NOR_PPI_AMD                
                        	nor_ops = &ppi_amd;
                        	break;
                        #else
	                        printk("NOR: not support ppi-amd flash type\n");
	                        return -1;                   
                        #endif
                case 2: /* ppi intel */
			#ifdef CONFIG_PL_NOR_PPI_INTEL                
                        	nor_ops = &ppi_intel;
                        	break;
                        #else
	                        printk("NOR: not support ppi-intel flash type\n");
	                        return -1;                   
                        #endif
                default:
                        printk("NOR: unknown flash type\n");
                        return -1;
        }
        nor_ops->init();
        nor_ops->read_id(&id);
        printk("NOR flash id = 0x%02x\n", id);

        return 0;
}

static int __init pl_nor_init_dma(void)
{
#if 1
	int rc;
#else
/* We won't need this for now. Use the buffer passed from MTD */
        int rc, size, order;
        int page_cnt[10] = { 1, 2, 4, 8, 16, 32, 64, 128, 256, 512 };

        if ( dma_buf_size == 0 ) {
                printk("NOR: dma_buf_size == 0? failed to initialize\n");
                return -1;
        }
        size = (dma_buf_size >> 12);
        for ( order=0; order<10; order++ ) {
                if ( page_cnt[order] == size )
                        break;
        }
        if ( order >= 10 ) {
                printk("NOR: page count = %d, not a happy value for me, failed init.\n", size);
                return -1;
        }
        /* Use __get_free_pages to get pages to be used in DMA transfer.
         * Note: because of page nature, it's guaranteed to be 16-byte
         * aligned(4K page aligned exactly), which is our controller's
         * requirement for DMA.
         */
        dma_buf = (char *)__get_free_pages(__GFP_DMA, order);
        if ( !dma_buf ) {
                printk("NOR: cannot get %d pages for DMA, failed init.\n", size);
                return -1;
        }
#endif
         
#if 0
        printk("NOR: allocate %d pages for DMA buffer\n", size);
#endif
        memset(iovec_tbl, 0x00, sizeof(struct iovec)*(MDE_MAX_ENTRY+1));
        rc = request_dma(DMA_PL_NOR, "NOR");
        if ( rc < 0 )
                return rc;
                
        set_dma_mde_base(DMA_PL_NOR, (const char *)mdt);
        set_dma_mode(DMA_PL_NOR, DMA_MODE_READ);
        set_dma_burst_length(DMA_PL_NOR, DMA_BURST_LINE);

        return 0;
}

static int __init pl_nor_init_irq(void)
{
        int rc;
        
        intr_status = readb(INTR_S_REG); /* in case there are still pending intr */
        
        rc = request_irq(IRQ_PL_NOR, pl_nor_do_interrupt, 0, "nor", NULL);
        if ( rc ) {
                printk("NOR: cannot register IRQ%d for block driver, rc=%d\n",
                       IRQ_PL_NOR, rc);
                return -1;
        } else {
                printk("nor interrupt %d registered\n", IRQ_PL_NOR);
        }

        /* mask three interrupts */
        NOR_INTR_SET_MASK(NOR_INTR_MASKALL);
        
        return rc;
}

static int __init pl_nor_init(void)
{
	pl_nor_mtd_banner();
#if 0
	if ( !norflash_tag ) {
		printk("No NOR flash tag passed by Proboot, failed to initialize\n");
		return -1;
	}
#endif
	pl_nor_init_controller();
	pl_nor_generate_map();
	
	if ( pl_nor_init_dma()<0 )
		return -1;
		
	if ( pl_nor_init_irq()<0 )
		return -1;
		
	return 0;
}

static int pl_nor_erase_one_block(unsigned long addr)
{
        int rc = 0;
        
#if USE_POLLING
	nor_ops->cmd_erase(addr, DMA_START);
	if(!pl_nor_polling_intr(NOR_CMD_INTR_MASK,0)) printk("pl_nor_erase_one_block polling failed?!\n");
        printk("CTRL_ST_REG = %x\n", readb(CTRL_ST_REG));
	
#else
        wait_queue_t wait_erase;
        NOR_INTR_SET_MASK(NOR_DMA_INTR_MASK);
        readb(INTR_S_REG); //read clear
        intr_status = 0x00;
        init_waitqueue_entry(&wait_erase, current);
        add_wait_queue(&wq_intr, &wait_erase);
        set_current_state(TASK_INTERRUPTIBLE);           
        nor_ops->cmd_erase(addr, DMA_START);
        schedule();
        set_current_state(TASK_RUNNING);
        remove_wait_queue(&wq_intr, &wait_erase);
        if ( (intr_status & 0x04) != 0 ) {
                printk("nor erase sector %lu failed?!\n", addr);
        }
#endif
        rc = nor_ops->cmd_erase(addr, DMA_END);
        if ( rc < 0 ) {
                printk("nor erase data polling error?!\n");
        }
        NOR_INTR_SET_MASK(NOR_INTR_MASKALL);
        
        return rc;
}

static int pl_nor_map_sector(unsigned long addr, nor_sector_t *sector)
{
        int i;
        int rc = -EINVAL;
        
        for ( i=0; i<total_sectors; i++ ) {
                if ( (addr >= nor_sector_map[i].start_addr) &&
                     (addr < nor_sector_map[i].start_addr + 
                             nor_sector_map[i].size )          
                             ) {
                        sector->start_addr = nor_sector_map[i].start_addr;
                        sector->size = nor_sector_map[i].size;
                        rc = 0;
                        break;
                }
        }
        
        return rc;
}

/* MTD related functions */

static struct mtd_info pl_nor_mtd;
static char pl_nor_mtd_name[] = "plnormtd";

static int pl_nor_flash_read(struct mtd_info *mtd, loff_t from, size_t len,
			     size_t *retlen, u_char *buf)
{
	int rc, pio_len = 0, this_cnt;
	u_char *buf_now;
	loff_t addr;
	
#if 0
printk("NOR read from %lld len = %d to buf %p\n", from, len, buf);
#endif

	rc = -EIO;
		
	if ( from+len > mtd->size )
		return -EINVAL;

        if ( down_interruptible(&ctl_sem) ) {
                return -ERESTARTSYS;
        }
        
        addr = from;
        buf_now = (u_char *)buf;
        *retlen = 0;
        if ( (unsigned long)buf_now & 0x0f ) {
        	pio_len = (unsigned long)buf_now & 0x0f;
		pio_len = len < pio_len ? len : pio_len;
#if 0
printk("NOR pio read: %lld\n", from);
#endif        	
        	nor_ops->pio_read(from, buf_now, pio_len);
        	buf_now += pio_len;
        	*retlen = pio_len;
        	len -= pio_len;
        	addr += pio_len;
        }

        rc = 0;
        while ( len > 0 ) {
        	this_cnt = len > MAX_DMA_CNT ? MAX_DMA_CNT : len;
        	if(len<16){
#if 0	
printk("NOR pio read %d bytes,", len);
#endif        	
			nor_ops->pio_read(addr, buf_now, len);  
	                buf_now += len;
	                addr += len;
	                *retlen += len;	
	                len -= len;	
#if 0	
printk(" retlen = %d bytes\n", *retlen);
#endif        		                      		
        	}
        	else{        
	        	if ( (unsigned long)this_cnt & 0x0f ) {
	        		this_cnt=this_cnt-( (unsigned int)this_cnt & 0x0f );
				//printk("NOR read this_cnt = %d bytes\n", this_cnt);
	        	}        		
	                iovec_tbl[0].iov_base = buf_now;
	                iovec_tbl[0].iov_len = this_cnt;
	                if ( pl_setup_dma(DMA_MODE_READ, iovec_tbl, 1) != this_cnt ) {
	                        printk("Cannot setup PL DMA channel for read\n");
	                        rc = -EFAULT;
	                        goto out;
	                }
	
#if USE_POLLING
	                enable_dma(DMA_PL_NOR);
	                nor_ops->cmd_read(addr, this_cnt, DMA_START);
			pl_nor_polling_intr(NOR_CMD_INTR_MASK,0);
	                if ( !(get_dma_residue(DMA_PL_NOR)==0) ) {
	                        printk("nor mtd dma read error?!\n");
	                        rc = -EIO;
	                }		
#else        
	        	wait_queue_t wait_dma;
	                intr_status = 0x00;
	                NOR_INTR_SET_MASK(NOR_CMD_INTR_MASK);
	                enable_dma(DMA_PL_NOR);
		        init_waitqueue_entry(&wait_dma, current);
		        add_wait_queue(&wq_intr, &wait_dma);
		        set_current_state(TASK_INTERRUPTIBLE);
	                nor_ops->cmd_read(addr, this_cnt, DMA_START);
	                schedule();
	                set_current_state(TASK_RUNNING);
		        remove_wait_queue(&wq_intr, &wait_dma);
	        
	                if ( !((intr_status & 0x01) == 0 && get_dma_residue(DMA_PL_NOR)==0) ) {
	                        printk("nor mtd dma read error?!\n");
	                        rc = -EIO;
	                }
#endif                
	                NOR_INTR_SET_MASK(NOR_INTR_MASKALL);
	                disable_dma(DMA_PL_NOR);
	                pl_unmap_dma(iovec_tbl, 1, DMA_MODE_READ);
	                nor_ops->cmd_read(addr, this_cnt, DMA_END);
	                buf_now += this_cnt;
	                len -= this_cnt;
	                addr += this_cnt;
	                *retlen += this_cnt;
	        }
        }
out:
        up(&ctl_sem);
        
        return rc;
}

static int pl_nor_flash_write(struct mtd_info *mtd, loff_t to, size_t len,
			      size_t *retlen, const u_char *buf)
{
	int rc, pio_len = 0, this_cnt;
	u_char *buf_now;
	loff_t addr;
	
#if 0
printk("NOR write to %lld len = %d from buf %p\n", to, len, buf);
#endif	
	rc = -EIO;
	
	if ( to+len > mtd->size )
		return -EINVAL;
	
        if ( down_interruptible(&ctl_sem) )
                return -ERESTARTSYS;

	addr = to;
	buf_now = (u_char *)buf;
	*retlen = 0;
	if ( (unsigned long)buf_now & 0x0f ) {
		pio_len = (unsigned long)buf_now & 0x0f;
		pio_len = len < pio_len ? len : pio_len;
#if 0
printk("NOR pio write: %lld\n", to);
#endif
		nor_ops->pio_write(to, buf_now, pio_len);
		buf_now += pio_len;
		*retlen = pio_len;
		len -= pio_len;
		addr += pio_len;
	}

	rc = 0;	
        while ( len > 0 ) {
        	this_cnt = len > MAX_DMA_CNT ? MAX_DMA_CNT : len;
        	if(len<16){
#if 0	
printk("NOR pio write %d bytes,", len);
#endif        	
			nor_ops->pio_write(addr, buf_now, len);  
	                buf_now += len;
	                addr += len;
	                *retlen += len;	
	                len -= len;	
#if 0	
printk(" retlen = %d bytes\n", *retlen);
#endif        		                      		
        	}
        	else{         
	        	if ( (unsigned long)this_cnt & 0x0f ) {
	        		this_cnt=this_cnt-( (unsigned int)this_cnt & 0x0f );
				//printk("NOR write this_cnt = %d bytes , from %p\n", this_cnt,buf_now);
	        	}      	
	        	iovec_tbl[0].iov_base = buf_now;
	                iovec_tbl[0].iov_len = this_cnt;
	                if ( pl_setup_dma(DMA_MODE_WRITE, iovec_tbl, 1) != this_cnt ) {
	                        printk("Cannot setup PL DMA channel for write\n");
	                        rc = -EFAULT;
	                        goto out;
	                }
	
#if USE_POLLING
	                enable_dma(DMA_PL_NOR);
	                nor_ops->cmd_write(addr, this_cnt, DMA_START);
			pl_nor_polling_intr(NOR_CMD_INTR_MASK,0);
	                if ( !(get_dma_residue(DMA_PL_NOR)==0) ) {
	                        printk("nor mtd dma read error?!\n");
	                        rc = -EIO;
	                }		
#else              
	        	wait_queue_t wait_dma;
	                intr_status = 0x00;
	                NOR_INTR_SET_MASK(NOR_CMD_INTR_MASK);
	                enable_dma(DMA_PL_NOR);
		        init_waitqueue_entry(&wait_dma, current);
		        add_wait_queue(&wq_intr, &wait_dma);
		        set_current_state(TASK_UNINTERRUPTIBLE);
	                nor_ops->cmd_write(addr, this_cnt, DMA_START);
	                schedule();
	                set_current_state(TASK_RUNNING);
		        remove_wait_queue(&wq_intr, &wait_dma);
	                if ( !((intr_status & 0x01) == 0 && get_dma_residue(DMA_PL_NOR)==0) ) {
	                        printk("nor mtd dma write error?!\n");
	                        rc = -EIO;
	                }
#endif                
	                NOR_INTR_SET_MASK(NOR_INTR_MASKALL);
	                disable_dma(DMA_PL_NOR);
	                pl_unmap_dma(iovec_tbl, 1, DMA_MODE_WRITE);
	                nor_ops->cmd_write(addr, this_cnt, DMA_END);
	                
	                buf_now += this_cnt;
	                len -= this_cnt;
	                addr += this_cnt;
	                *retlen += this_cnt;
	        }
        }
out:
        up(&ctl_sem);
        
        return rc;

}

static int pl_nor_flash_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	int len;
	unsigned long addr;
	int rc = -EINVAL;
#if 0
printk("nor mtd erase, addr = %u len = %u\n", instr->addr, instr->len);
#endif

	instr->state = MTD_ERASE_FAILED;

	if ( instr->addr > mtd->size ) 
		return -EINVAL;
	if ( instr->addr + instr->len > mtd->size )
		return -EINVAL;
#if 0
	/* 
	 * Do we need to check if the erase starting address and length
	 * are erase-block aligned? 
	 */
#endif		

	len = instr->len;
	addr = instr->addr;
	
	if ( down_interruptible(&ctl_sem) )
		return -ERESTARTSYS;
	
	while ( len > 0 ) {
		nor_sector_t sector;
		
		if ( pl_nor_map_sector(addr, &sector) < 0 ) {
			printk("nor addr %lu cannot be mapped to sector\n", addr);
			goto probably_erase_failed;
		}
#if 0
printk("nor mtd erase %lu block\n", sector.start_addr);
#endif
		if ( pl_nor_erase_one_block(sector.start_addr) < 0 ) {
			printk("nor erase block addr %lu failed\n", sector.start_addr);
			goto probably_erase_failed;
		}
		len -= sector.size;
		addr += sector.size;
	}
	
	rc = 0;
	instr->state = MTD_ERASE_DONE;
	if ( instr->callback )
		instr->callback(instr);
	
probably_erase_failed:	
	up(&ctl_sem);
		
	return rc;
}

static struct mtd_partition pl_nor_partitions[] = {
#if USE_PPI_EN29LV320B	
	{
		name: 		"bootsect",
		size: 		8192,
		offset: 	0,
		mask_flags:	MTD_WRITEABLE, /* read-only */
	}, {
		name:		"autoboot.bat",
		size:		8192,
		offset:		MTDPART_OFS_APPEND, /* starts at 0x2000 */
	}, {
		name:		"registry",
		size:		0xc000, /* 6*8K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x4000 */
	}, {
		name:		"proboot",
		size:		0x10000, /* 64K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x10000 */
	}, {
		name:		"kernel",
		size:		0xd0000, /* 13*64K = 832K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x20000 */
	}, {
		name:		"rootfs",
		size:		0x250000, /* 37*64K = 2368K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0xf0000 */
	}, {
		name:		"reserved",
		size:		MTDPART_SIZ_FULL,
		offset:		MTDPART_OFS_APPEND, /* starts at 0x340000 */
	}, { /* this is overlapped with kernel and rootfs above */
		name:		"system",
		size:		0x1e0000,
		offset:		0x20000,
	}
#elif USE_SPI_S25FL032A
	{
		name: 		"bootsect",
		size: 		0x10000,
		offset: 	0,
		mask_flags:	MTD_WRITEABLE, /* read-only */
	}, {
		name:		"proboot",
		size:		0x10000,
		offset:		MTDPART_OFS_APPEND, /* starts at 0x10000 */
	}, {
		name:		"autoboot.bat",
		size:		0x20000, /* 2*64K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x20000 */
	}, {
		name:		"registry",
		size:		0x10000, /* 64K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x40000 */
	}, {
		name:		"kernel",
		size:		0xc0000, /* 12*64K = 768K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x50000 */
	}, {
		name:		"rootfs",
		size:		0x200000, /* 32*64K = 2048K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x110000 */
	}, {
		name:		"reserved",
		size:		MTDPART_SIZ_FULL,
		offset:		MTDPART_OFS_APPEND, /* starts at 0x310000 */
	}
#elif USE_PPI_INTEL_P30
	{
		name: 		"bootsect",
		size: 		0x8000,
		offset: 	0,
		mask_flags:	MTD_WRITEABLE, /* read-only */
	}, {
		name:		"autoboot.bat",
		size:		0x8000,
		offset:		MTDPART_OFS_APPEND, /* starts at 0x8000 */
	}, {
		name:		"registry",
		size:		0x10000, /* 2*32K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x10000 */
	}, {
		name:		"proboot",
		size:		0x20000, /* 128K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x20000 */
	}, {
		name:		"kernel",
		size:		0xc0000, /* 12*64K = 768K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x40000 */
	}, {
		name:		"rootfs",
		size:		0x200000, /* 32*64K = 2048K */
		offset:		MTDPART_OFS_APPEND, /* starts at 0x100000 */
	}, {
		name:		"reserved",
		size:		MTDPART_SIZ_FULL,
		offset:		MTDPART_OFS_APPEND, /* starts at 0x300000 */
	}
#endif	
};

static int initialized = 0;

static void cleanup_mtd_struct(void)
{
	if ( pl_nor_mtd.eraseregions != NULL )
		kfree(pl_nor_mtd.eraseregions);
	memset(&pl_nor_mtd, 0x00, sizeof(struct mtd_info));
}

int __init pl_nor_mtd_init(void)
{
	int i, rc;
	
	if ( initialized ) {
		printk("NOR MTD already initialized?");
		return -1;
	}
	
	rc = pl_nor_init();
	if ( rc < 0 )
		return rc;
	
	memset(&pl_nor_mtd, 0x00, sizeof(struct mtd_info));
	pl_nor_mtd.unpoint = NULL;
	pl_nor_mtd.type = MTD_NORFLASH;
	pl_nor_mtd.flags = MTD_CAP_NORFLASH;
	pl_nor_mtd.size = total_size;
	pl_nor_mtd.name = pl_nor_mtd_name;
	pl_nor_mtd.erasesize = 0; /* would update accordingly later */
	pl_nor_mtd.numeraseregions = num_erase_regions;
	pl_nor_mtd.eraseregions = (struct mtd_erase_region_info *)kmalloc(
				   sizeof(struct mtd_erase_region_info) *
				   num_erase_regions, GFP_KERNEL);
	if ( pl_nor_mtd.eraseregions == NULL ) {
		printk("Failed to allocate MTD erase regions info\n");
		kfree(pl_nor_mtd.name);
		return -1;
	}

do { /* setup MTD erase region */
	int c_size, c_blk, c_idx;
	struct mtd_erase_region_info *p;
	
	c_size = c_blk = 0;
	c_idx = -1; /* make sure we can be right at the first time */
	p = pl_nor_mtd.eraseregions;
	for ( i=0; i<total_sectors; i++ ) {
		if ( nor_sector_map[i].size != c_size ) {
			c_idx ++;
			if ( c_idx >= num_erase_regions ) {
			/* shouldn't be here */
				printk("Shouldn't be here... Wrong!!");
				break;
			}
			p = &(pl_nor_mtd.eraseregions[c_idx]);
			p->offset = nor_sector_map[i].start_addr;
			p->erasesize = c_size = nor_sector_map[i].size;
			p->numblocks = 1;
			
			if ( pl_nor_mtd.erasesize < p->erasesize )
				pl_nor_mtd.erasesize = p->erasesize;
			
			continue;
		}
		p->numblocks ++;
	}
#if 0
	printk("NOR MTD erase regions: \n");
	for ( i=0; i<num_erase_regions; i++ ) {
		struct mtd_erase_region_info *p;
			
		p = &(pl_nor_mtd.eraseregions[i]);
		printk("  Start %u erase size %u total blocks %u\n",
		       p->offset, p->erasesize, p->numblocks);
	}
#endif		
} while (0); 
	
	pl_nor_mtd.module = THIS_MODULE;
	pl_nor_mtd.erase = pl_nor_flash_erase;
	pl_nor_mtd.read = pl_nor_flash_read;
	pl_nor_mtd.write = pl_nor_flash_write;
	
/*
	rc = add_mtd_device(&pl_nor_mtd);
	if ( rc < 0 ) {
		printk("Cannot register NOR MTD, rc = %d\n", rc);
		goto mtd_init_failed;
	}
*/	
	rc = add_mtd_partitions(&pl_nor_mtd, pl_nor_partitions, 
			 	ARRAY_SIZE(pl_nor_partitions));
	if ( rc < 0 ) {
		printk("Cannot add mtd partitions, rc = %d\n", rc);
		goto mtd_init_failed;
	}
	
	initialized = 1;
	
	return 0;
	
mtd_init_failed:
	cleanup_mtd_struct();
	return -1;	
}

void __exit pl_nor_mtd_exit(void)
{
	if ( initialized == 0 )
		return;
		
	cleanup_mtd_struct();
}

static int __init pl_nor_parse_map(char *line)
{
	printk("\npl_nor_parse_map(%s)\n",line);
}
__setup("nor=", pl_nor_parse_map);

EXPORT_SYMBOL(pl_nor_polling_intr);

module_init(pl_nor_mtd_init);
module_exit(pl_nor_mtd_exit);

